// Dessembler by wiiqwertyuiop
// Usage: Dessembler <Patched ROM> <Clean ROM> <Patch> <Asar or Xkas>
//
// Here is what it does:
// It creates a temp file the same size as the patched ROM and patches the patch to it
// It then searches in the non-freespace area in the ROM and looks for changes by
// checking if the data in the temp file matches the data in the dirty ROM and checking if it doesn't match in the clean ROM
// if it finds changed data it changes it back to the clean ROM's value and continues looping
// after it is done with the non-freespace area it checks the temp file for a RATS tag in the freespace area
// if it finds one it looks for the exact matching data in the dirty ROM and when/if it finds it changes it to 00s and continues looping until the end
// however if it also finds data that is unprotected it guesses by removing the data in the ROM at that spot; i am not exactly sure if this is a good idea...
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

const char* get_ext(char* file){	// Get file extension
	const char* dot = strrchr(file, '.');
	if(!dot || dot == file) return "";
    return dot;
}

int ROMsize;
char * ROMdata;
char * tempData;

int main(int argc, char* argv[]) {

	char* ROMname;
	char* ROM2name;
	char* Patch;
	char assembler[] = "asar";

	printf("\Dessembler by wiiqwertyuiop\n\n");

	if(argc<4) {
		ROMname=(char*)malloc(256);
		printf("Patched ROM: ");
		gets(ROMname);
		ROM2name=(char*)malloc(256);
		printf("Clean ROM: ");
		gets(ROM2name);
		Patch=(char*)malloc(256);
		printf("Patch to remove: ");
		gets(Patch);
	} else {
		ROMname = argv[1];
		ROM2name = argv[2];
		Patch = argv[3];
	}

	char TempName[14] = "tempROM";
	strcat(TempName, get_ext(ROMname)); // get ROM extension
	
	FILE * ROM = fopen(ROMname, "r+b");	// open patched ROM
	if (ROM==NULL) {
		printf("Error opening: %s\n", ROMname);
		return 1;
	}

	fseek(ROM, 0, SEEK_END);		// get patched ROM size
	ROMsize=ftell(ROM);

	if(ROMsize == 0) {
		printf("Error %s is empty.\n", ROMname);
		return 1;
	}
	rewind(ROM);

	// allocate memory to contain the whole file:
	ROMdata = (char*) malloc (sizeof(char)*ROMsize);
	if (ROMdata == NULL) {
		printf("Memory error.\n");
		return 1;
	}
	// copy the file into the buffer:
	if (fread(ROMdata,1,ROMsize,ROM) != ROMsize) {
		 printf("Reading error.\n"); 
		 return 1;
	}
	rewind(ROM);



	FILE * CleanROM = fopen(ROM2name, "rb");	// open clean ROM

	if (CleanROM==NULL) {
		printf("Error opening: %s\n", ROM2name);
		return 1;
	}

	fseek(CleanROM, 0, SEEK_END);		// get clean ROM size
	int i=ftell(CleanROM);

	if(i == 0) {
		printf("Error %s is empty.\n", ROMname);
		return 1;
	}
	rewind(CleanROM);

	// allocate memory to contain the whole file:
	char * CleanData = (char*) malloc (sizeof(char)*i);
	if (CleanData == NULL) {
		printf("Memory error.\n");
		return 1;
	}
	// copy the file into the buffer:
	if (fread(CleanData,1,i,CleanROM) != i) {
		 printf("Reading error.\n"); 
		 return 1;
	}
	fclose(CleanROM);

	FILE * temp = fopen(TempName, "w+b");	// create temp ROM
	if(temp==NULL) {
		printf("Error couldn't make temp file.\n");
		return 1;
	}

	fseek(temp, ROMsize-1, SEEK_SET);	// make it the same size as the patched ROM
	fputc(0, temp);
	rewind(temp);


	char cmdline[512];

	// bah this feels so messy...		
#ifdef _WIN32
	if(rename("xkas.exe", "xkas.exe") == 0) {	// Cheap way of check if a file exists

		if(rename("asar.exe", "asar.exe") == 0) {
			printf("Use asar? y or n:\n");
			if(getchar() == 'y') goto asar;
		}
		assembler[0] = 'x';
		assembler[1] = 'k';
		assembler[2] = 'a';
		assembler[3] = 's';
		sprintf(cmdline, "%s %s %s", assembler, Patch, TempName);
		goto xkas;
	}
asar:
	sprintf(cmdline, "%s -nocheck %s %s", assembler, Patch, TempName);
#else
	if(rename("xkas", "xkas") == 0) {	// Cheap way of check if a file exists

		if(rename("asar", "asar") == 0) {
			printf("Use asar? y or n:\n");
			if(getchar() == 'y') goto asar;
		}
		assembler[0] = 'x';
		assembler[1] = 'k';
		assembler[2] = 'a';
		assembler[3] = 's';
		sprintf(cmdline, "%s %s %s", assembler, Patch, TempName);
		goto xkas;
	}
asar:
	sprintf(cmdline, "./%s -nocheck %s %s", assembler, Patch, TempName);
#endif
xkas:

	// patch the patch to the temp file
	if(system(cmdline) != 0) return 1;

	// allocate memory to contain the whole file:
	tempData = (char*) malloc (sizeof(char)*ROMsize);
	if (tempData == NULL) {
		printf("Memory error.\n");
		return 1;
	}
	// copy the file into the buffer:
	if (fread(tempData,1,ROMsize,temp) != ROMsize) {
		 printf("Reading error.\n"); 
		 return 1;
	}
	fclose(temp);

	int a = 0;		// loop through clean file
	while(a<i) {

		if( (tempData[a] == ROMdata[a]) && (tempData[a] != CleanData[a]) ) {	// find differences between the patched ROM and clean ROM
			fseek(ROM, a, SEEK_SET);
			fputc(CleanData[a], ROM);		// change it to the orginal data
		}
		a++;
	}

	free(CleanData);
	fseek(ROM, a, SEEK_SET);

	while(a<ROMsize) {	// continue looping through the patched ROM

		if(tempData[a] == 'S' && tempData[a+1] == 'T' && tempData[a+2] == 'A' && tempData[a+3] == 'R') {		// we should now be in freespace area so look for a RATS tag
			a += 4;
			int datasize = tempData[a]|(tempData[a+1]<<8);	// get size of data from tag
			//a += 4;		// match data size as well

			int clear = -1;
			i = 0;
			while(i<ROMsize) {

				if(ROMdata[i] == tempData[a]) {		// look for matching data in the ROM
					int x = 1;
					clear = i;
					while(x < datasize+1) {		// see if all of the data matches
						if(ROMdata[i+x] != tempData[a+x]) {
							clear = -1;
						}
						x++;
					}

					if(clear != -1) {	
						// if we have a match
						int x = 0;

						fseek(ROM, i-4, SEEK_SET);
						while(x<datasize+8+1) {		// clear data
							fputc(0, ROM);
							x++;
						}
						break;
					}
				}
				i++;
			}

		} else if(tempData[a] != 0) {	// If we find data not protected by a tag get rid of it
			fseek(ROM, a, SEEK_SET);
			fputc(0, ROM);
		}
		a++;
	}

	fclose(ROM);
	free(tempData);
	free(ROMdata);

	if(remove(TempName) != 0) printf("Couldn't remove temp file.\n");

	printf("Done!\n");
	return 0;
}